@@ -1,23 +1,31 @@
 use crate::*;
 
-use crate::block::{Block, BlockPtr, Item, ItemContent, ID};
+use crate::block::{Block, BlockPtr, Item, ItemContent, Prelim, ID};
 use crate::block_store::StateVector;
 use crate::event::UpdateEvent;
 use crate::id_set::{DeleteSet, IdSet};
 use crate::store::Store;
 use crate::types::array::Array;
-use crate::types::{Map, Text, TypePtr, TYPE_REFS_ARRAY, TYPE_REFS_MAP, TYPE_REFS_TEXT};
+use crate::types::xml::{XmlElement, XmlText};
+use crate::types::{
+    Branch, Map, Text, TypePtr, TYPE_REFS_ARRAY, TYPE_REFS_MAP, TYPE_REFS_TEXT,
+    TYPE_REFS_XML_ELEMENT, TYPE_REFS_XML_TEXT,
+};
 use crate::update::Update;
 use std::cell::RefMut;
 use std::collections::{HashMap, HashSet};
 use std::ops::Range;
 use updates::encoder::*;
 
+/// Transaction is one of the core types in Yrs. All operations that need to touch a document's
+/// contents (a.k.a. block store), need to be executed in scope of a transaction.
 pub struct Transaction<'a> {
     /// Store containing the state of the document.
-    pub store: RefMut<'a, Store>,
-    /// State vector of a current transaction.
+    pub(crate) store: RefMut<'a, Store>,
+    /// State vector of a current transaction at the moment of its creation.
     pub before_state: StateVector,
+    /// Current state vector of a transaction, which includes all performed updates.
+    pub after_state: StateVector,
     /// ID's of the blocks to be merged.
     pub merge_blocks: Vec<ID>,
     /// Describes the set of deleted items by ids.
@@ -25,11 +33,10 @@ pub struct Transaction<'a> {
     /// All types that were directly modified (property added or child inserted/deleted).
     /// New types are not included in this Set.
     changed: HashMap<TypePtr, HashSet<Option<String>>>,
-    after_state: StateVector,
 }
 
 impl<'a> Transaction<'a> {
-    pub fn new(store: RefMut<'a, Store>) -> Transaction {
+    pub(crate) fn new(store: RefMut<'a, Store>) -> Transaction {
         let begin_timestamp = store.blocks.get_state_vector();
         Transaction {
             store,
@@ -41,19 +48,95 @@ impl<'a> Transaction<'a> {
         }
     }
 
+    /// Returns state vector describing current state of the updates.
+    pub fn state_vector(&self) -> StateVector {
+        self.store.blocks.get_state_vector()
+    }
+
+    /// Encodes the difference between remove peer state given its `state_vector` and the state
+    /// of a current local peer
+    pub fn encode_diff<E: Encoder>(&self, state_vector: &StateVector, encoder: &mut E) {
+        self.store.encode_diff(state_vector, encoder)
+    }
+
+    /// Returns a [Text] data structure stored under a given `name`. Text structures are used for
+    /// collaborative text editing: they expose operations to append and remove chunks of text,
+    /// which are free to execute concurrently by multiple peers over remote boundaries.
+    ///
+    /// If not structure under defined `name` existed before, it will be created and returned
+    /// instead.
+    ///
+    /// If a structure under defined `name` already existed, but its type was different it will be
+    /// reinterpreted as a text (in such case a sequence component of complex data type will be
+    /// interpreted as a list of text chunks).
     pub fn get_text(&mut self, name: &str) -> Text {
-        let c = self.store.create_type(name, TYPE_REFS_TEXT);
-        Text::new(c)
+        let c = self.store.create_type(name, None, TYPE_REFS_TEXT);
+        Text::from(c)
     }
 
+    /// Returns a [Map] data structure stored under a given `name`. Maps are used to store key-value
+    /// pairs associated together. These values can be primitive data (similar but not limited to
+    /// a JavaScript Object Notation) as well as other shared types (Yrs maps, arrays, text
+    /// structures etc.), enabling to construct a complex recursive tree structures.
+    ///
+    /// If not structure under defined `name` existed before, it will be created and returned
+    /// instead.
+    ///
+    /// If a structure under defined `name` already existed, but its type was different it will be
+    /// reinterpreted as a map (in such case a map component of complex data type will be
+    /// interpreted as native map).
     pub fn get_map(&mut self, name: &str) -> Map {
-        let c = self.store.create_type(name, TYPE_REFS_MAP);
-        Map::new(c)
+        let c = self.store.create_type(name, None, TYPE_REFS_MAP);
+        Map::from(c)
     }
 
+    /// Returns an [Array] data structure stored under a given `name`. Array structures are used for
+    /// storing a sequences of elements in ordered manner, positioning given element accordingly
+    /// to its index.
+    ///
+    /// If not structure under defined `name` existed before, it will be created and returned
+    /// instead.
+    ///
+    /// If a structure under defined `name` already existed, but its type was different it will be
+    /// reinterpreted as an array (in such case a sequence component of complex data type will be
+    /// interpreted as a list of inserted values).
     pub fn get_array(&mut self, name: &str) -> Array {
-        let c = self.store.create_type(name, TYPE_REFS_ARRAY);
-        Array::new(c)
+        let c = self.store.create_type(name, None, TYPE_REFS_ARRAY);
+        Array::from(c)
+    }
+
+    /// Returns a [XmlElement] data structure stored under a given `name`. XML elements represent
+    /// nodes of XML document. They can contain attributes (key-value pairs, both of string type)
+    /// as well as other nested XML elements or text values, which are stored in their insertion
+    /// order.
+    ///
+    /// If not structure under defined `name` existed before, it will be created and returned
+    /// instead.
+    ///
+    /// If a structure under defined `name` already existed, but its type was different it will be
+    /// reinterpreted as a XML element (in such case a map component of complex data type will be
+    /// interpreted as map of its attributes, while a sequence component - as a list of its child
+    /// XML nodes).
+    pub fn get_xml_element(&mut self, name: &str) -> XmlElement {
+        let c = self
+            .store
+            .create_type(name, Some("UNDEFINED".to_string()), TYPE_REFS_XML_ELEMENT);
+        XmlElement::from(c)
+    }
+
+    /// Returns a [XmlText] data structure stored under a given `name`. Text structures are used for
+    /// collaborative text editing: they expose operations to append and remove chunks of text,
+    /// which are free to execute concurrently by multiple peers over remote boundaries.
+    ///
+    /// If not structure under defined `name` existed before, it will be created and returned
+    /// instead.
+    ///
+    /// If a structure under defined `name` already existed, but its type was different it will be
+    /// reinterpreted as a text (in such case a sequence component of complex data type will be
+    /// interpreted as a list of text chunks).
+    pub fn get_xml_text(&mut self, name: &str) -> XmlText {
+        let c = self.store.create_type(name, None, TYPE_REFS_XML_TEXT);
+        XmlText::from(c)
     }
 
     /// Encodes the document state to a binary format.
@@ -64,14 +147,14 @@ impl<'a> Transaction<'a> {
     ///   end up with the same content.
     /// * Even if an update contains known information, the unknown information
     ///   is extracted and integrated into the document structure.
-    pub fn encode_update(&self) -> Vec<u8> {
+    pub fn encode_update_v1(&self) -> Vec<u8> {
         let mut update_encoder = updates::encoder::EncoderV1::new();
         self.store
             .encode_diff(&self.before_state, &mut update_encoder);
         update_encoder.to_vec()
     }
 
-    pub fn iterate_structs<F>(&mut self, client: &u64, range: &Range<u32>, f: &F)
+    pub(crate) fn iterate_structs<F>(&mut self, client: &u64, range: &Range<u32>, f: &F)
     where
         F: Fn(&Block) -> (),
     {
@@ -101,10 +184,10 @@ impl<'a> Transaction<'a> {
         }
     }
 
-    pub fn find_index_clean_start(&mut self, client: &u64, clock: u32) -> Option<usize> {
+    pub(crate) fn find_index_clean_start(&mut self, client: &u64, clock: u32) -> Option<usize> {
         let blocks = self.store.blocks.get_mut(client)?;
         let index = blocks.find_pivot(clock)?;
-        let block = &mut blocks[index];
+        let block = blocks.get_mut(index);
         if let Some(item) = block.as_item_mut() {
             if item.id.clock < clock {
                 // if we run over the clock, we need to the split item
@@ -119,7 +202,7 @@ impl<'a> Transaction<'a> {
         Some(index)
     }
 
-    pub fn apply_ranges<F>(&mut self, set: &IdSet, f: &F)
+    pub(crate) fn apply_ranges<F>(&mut self, set: &IdSet, f: &F)
     where
         F: Fn(&Block) -> (),
     {
@@ -135,7 +218,7 @@ impl<'a> Transaction<'a> {
 
     /// Applies given `id_set` onto current transaction to run multi-range deletion.
     /// Returns a remaining of original ID set, that couldn't be applied.
-    pub fn apply_delete(&mut self, ds: &DeleteSet) -> Option<DeleteSet> {
+    pub(crate) fn apply_delete(&mut self, ds: &DeleteSet) -> Option<DeleteSet> {
         let mut unapplied = DeleteSet::new();
         for (client, ranges) in ds.iter() {
             let mut blocks = self.store.blocks.get_mut(client).unwrap();
@@ -152,13 +235,11 @@ impl<'a> Transaction<'a> {
                     // We can ignore the case of GC and Delete structs, because we are going to skip them
                     if let Some(mut index) = blocks.find_pivot(clock) {
                         // We can ignore the case of GC and Delete structs, because we are going to skip them
-                        if let Some(item) = blocks[index].as_item_mut() {
+                        if let Some(item) = blocks.get_mut(index).as_item_mut() {
                             // split the first item if necessary
                             if !item.is_deleted() && item.id.clock < clock {
-                                let split_ptr = BlockPtr::new(
-                                    ID::new(*client, clock - item.id.clock),
-                                    index as u32,
-                                );
+                                let split_ptr =
+                                    BlockPtr::new(ID::new(*client, clock), index as u32);
                                 let (_, right) = self.store.blocks.split_block(&split_ptr);
                                 if let Some(right) = right {
                                     index += 1;
@@ -168,7 +249,7 @@ impl<'a> Transaction<'a> {
                             }
 
                             while index < blocks.len() {
-                                let block = &mut blocks[index];
+                                let block = blocks.get_mut(index);
                                 if let Some(item) = block.as_item_mut() {
                                     if item.id.clock < clock_end {
                                         if !item.is_deleted() {
@@ -216,7 +297,7 @@ impl<'a> Transaction<'a> {
         let mut recurse = Vec::new();
         let mut result = false;
 
-        if let Some(item) = self.store.blocks.get_item(&ptr) {
+        if let Some(item) = self.store.blocks.get_item_mut(&ptr) {
             if !item.is_deleted() {
                 if item.parent_sub.is_none() && item.is_countable() {
                     if let Some(parent) = self.store.get_type(&item.parent) {
@@ -227,7 +308,25 @@ impl<'a> Transaction<'a> {
 
                 item.mark_as_deleted();
                 self.delete_set.insert(item.id.clone(), item.len());
-                // addChangedTypeToTransaction(transaction, item.type, item.parentSub)
+
+                match &item.parent {
+                    TypePtr::Named(_) => {
+                        self.changed
+                            .entry(item.parent.clone())
+                            .or_default()
+                            .insert(item.parent_sub.clone());
+                    }
+                    TypePtr::Id(ptr)
+                        if ptr.id.clock < self.before_state.get(&ptr.id.client)
+                            && self.store.blocks.get_item(ptr).unwrap().is_deleted() =>
+                    {
+                        self.changed
+                            .entry(item.parent.clone())
+                            .or_default()
+                            .insert(item.parent_sub.clone());
+                    }
+                    _ => {}
+                }
                 if item.id.clock < self.before_state.get(&item.id.client) {
                     let set = self.changed.entry(item.parent.clone()).or_default();
                     set.insert(item.parent_sub.clone());
@@ -244,7 +343,7 @@ impl<'a> Transaction<'a> {
                     }
                     ItemContent::Type(t) => {
                         let inner = t.borrow_mut();
-                        let mut ptr = inner.start.get();
+                        let mut ptr = inner.start;
                         self.changed.remove(&item.parent);
 
                         while let Some(item) = ptr.and_then(|ptr| self.store.blocks.get_item(&ptr))
@@ -260,9 +359,7 @@ impl<'a> Transaction<'a> {
                             recurse.push(ptr.clone());
                         }
                     }
-                    _ => {
-                        // nothing to do for other content types
-                    }
+                    _ => { /* nothing to do for other content types */ }
                 }
                 result = true;
             }
@@ -270,6 +367,10 @@ impl<'a> Transaction<'a> {
 
         for ptr in recurse.iter() {
             if !self.delete(ptr) {
+                // Whis will be gc'd later and we want to merge it if possible
+                // We try to merge all deleted items after each transaction,
+                // but we have no knowledge about that this needs to be merged
+                // since it is not in transaction.ds. Hence we add it to transaction._mergeStructs
                 self.merge_blocks.push(ptr.id);
             }
         }
@@ -277,14 +378,13 @@ impl<'a> Transaction<'a> {
         result
     }
 
-    pub fn apply_update(&mut self, mut update: Update, mut ds: DeleteSet) {
+    pub fn apply_update(&mut self, mut update: Update) {
         if self.store.update_events.has_subscribers() {
-            let event = UpdateEvent::new(update, ds);
+            let event = UpdateEvent::new(update);
             self.store.update_events.publish(&event);
             update = event.update;
-            ds = event.delete_set;
         }
-        let remaining = update.integrate(self);
+        let (remaining, remaining_ds) = update.integrate(self);
 
         let mut retry = false;
         if let Some(mut pending) = self.store.pending.take() {
@@ -301,43 +401,46 @@ impl<'a> Transaction<'a> {
                 for (&client, &clock) in remaining.missing.iter() {
                     pending.missing.set_min(client, clock);
                 }
-                pending.update.merge(remaining.update);
+                pending.update = Update::merge_updates(vec![pending.update, remaining.update]);
                 self.store.pending = Some(pending);
             }
         } else {
             self.store.pending = remaining;
         }
 
-        let ds = self.apply_delete(&ds);
         if let Some(pending) = self.store.pending_ds.take() {
             let ds2 = self.apply_delete(&pending);
-            let ds = match (ds, ds2) {
+            let ds = match (remaining_ds, ds2) {
                 (Some(mut a), Some(b)) => {
-                    a.merge(b);
-                    Some(a)
+                    a.delete_set.merge(b);
+                    Some(a.delete_set)
                 }
-                (Some(x), _) | (_, Some(x)) => Some(x),
+                (Some(x), _) => Some(x.delete_set),
+                (_, Some(x)) => Some(x),
                 _ => None,
             };
             self.store.pending_ds = ds;
         } else {
-            self.store.pending_ds = ds;
+            self.store.pending_ds = remaining_ds.map(|update| update.delete_set);
         }
 
         if retry {
             if let Some(pending) = self.store.pending.take() {
                 let ds = self.store.pending_ds.take().unwrap_or_default();
-                self.apply_update(pending.update, ds);
+                let mut ds_update = Update::new();
+                ds_update.delete_set = ds;
+                self.apply_update(pending.update);
+                self.apply_update(ds_update)
             }
         }
     }
 
-    pub fn create_item(
+    pub(crate) fn create_item<T: Prelim>(
         &mut self,
         pos: &block::ItemPosition,
-        content: block::ItemContent,
+        value: T,
         parent_sub: Option<String>,
-    ) {
+    ) -> &Item {
         let left = pos.left;
         let right = pos.right;
         let origin = if let Some(ptr) = pos.left.as_ref() {
@@ -360,6 +463,13 @@ impl<'a> Transaction<'a> {
             .get_client_blocks_mut(client_id)
             .integrated_len() as u32;
 
+        let ptr = BlockPtr::new(id, pivot);
+        let (content, remainder) = value.into_content(self, TypePtr::Id(ptr));
+        let inner_ref = if let ItemContent::Type(inner_ref) = &content {
+            Some(inner_ref.clone())
+        } else {
+            None
+        };
         let mut item = Item::new(
             id,
             left,
@@ -370,14 +480,32 @@ impl<'a> Transaction<'a> {
             parent_sub,
             content,
         );
+
         item.integrate(self, pivot, 0);
+
         let local_block_list = self.store.blocks.get_client_blocks_mut(client_id);
         local_block_list.push(block::Block::Item(item));
+
+        let idx = local_block_list.len() - 1;
+
+        if let Some(remainder) = remainder {
+            remainder.integrate(self, inner_ref.unwrap())
+        }
+
+        self.store.blocks.get_client_blocks_mut(client_id)[idx]
+            .as_item()
+            .unwrap()
     }
 
+    /// Commits current transaction. This step involves cleaning up and optimizing changes performed
+    /// during lifetime of a transaction. Such changes include squashing delete sets data
+    /// or squashing blocks that have been appended one after another to preserve memory.
+    ///
+    /// This step is performed automatically when a transaction is about to be dropped (its life
+    /// scope comes to an end).
     pub fn commit(&mut self) {
         // 1. sort and merge delete set
-        self.delete_set.compact();
+        self.delete_set.squash();
         self.after_state = self.store.blocks.get_state_vector();
 
         // 2. emit 'beforeObserverCalls'
@@ -386,7 +514,7 @@ impl<'a> Transaction<'a> {
         self.try_gc(); //TODO: eventually this is a configurable variant: if (doc.gc)
 
         // 5. try merge delete set
-        self.delete_set.try_compact(&self.store.blocks);
+        self.delete_set.try_squash_with(&mut self.store);
 
         // 6. get transaction after state and try to merge to left
         for (client, &clock) in self.after_state.iter() {
@@ -396,7 +524,7 @@ impl<'a> Transaction<'a> {
                 let first_change = blocks.find_pivot(before_clock).unwrap().max(1);
                 let mut i = blocks.len() - 1;
                 while i >= first_change {
-                    if let Some(compaction) = blocks.compact_left(i) {
+                    if let Some(compaction) = blocks.squash_left(i) {
                         self.store.gc_cleanup(compaction);
                         blocks = self.store.blocks.get_mut(client).unwrap();
                     }
@@ -411,11 +539,11 @@ impl<'a> Transaction<'a> {
             let blocks = self.store.blocks.get_mut(&client).unwrap();
             let replaced_pos = blocks.find_pivot(clock).unwrap();
             if replaced_pos + 1 < blocks.len() {
-                if let Some(compaction) = blocks.compact_left(replaced_pos + 1) {
+                if let Some(compaction) = blocks.squash_left(replaced_pos + 1) {
                     self.store.gc_cleanup(compaction);
                 }
             } else if replaced_pos > 0 {
-                if let Some(compaction) = blocks.compact_left(replaced_pos) {
+                if let Some(compaction) = blocks.squash_left(replaced_pos) {
                     self.store.gc_cleanup(compaction);
                 }
             }
@@ -427,43 +555,20 @@ impl<'a> Transaction<'a> {
         // 12. emit 'subdocs'
     }
 
-    fn try_gc(&mut self) {
+    fn try_gc(&self) {
         for (client, range) in self.delete_set.iter() {
-            if let Some(blocks) = self.store.blocks.get_mut(client) {
+            if let Some(blocks) = self.store.blocks.get(client) {
                 for delete_item in range.iter().rev() {
                     let mut start = delete_item.start;
                     if let Some(mut i) = blocks.find_pivot(start) {
                         while i < blocks.len() {
-                            let block = &mut blocks[i];
+                            let block = blocks.get_mut(i);
                             let len = block.len();
                             start += len;
                             if start > delete_item.end {
                                 break;
                             } else {
-                                if let Block::Item(item) = block {
-                                    if item.is_deleted() {
-                                        if let ItemContent::Type(t) = &item.content {
-                                            /*
-                                            let item = this.type._start
-                                            while (item !== null) {
-                                              item.gc(store, true)
-                                              item = item.right
-                                            }
-                                            this.type._start = null
-                                            this.type._map.forEach(/** @param {Item | null} item */ (item) => {
-                                              while (item !== null) {
-                                                item.gc(store, true)
-                                                item = item.left
-                                              }
-                                            })
-                                            this.type._map = new Map()
-                                            */
-                                            todo!()
-                                        }
-
-                                        item.content = ItemContent::Deleted(len);
-                                    }
-                                }
+                                block.gc(self, false);
                                 i += 1;
                             }
                         }
@@ -472,6 +577,24 @@ impl<'a> Transaction<'a> {
             }
         }
     }
+
+    pub(crate) fn add_changed_type(&mut self, parent: &Branch, parent_sub: Option<&String>) {
+        let trigger = match parent.item.as_ref() {
+            None => true,
+            Some(ptr) if ptr.id.clock < (self.before_state.get(&ptr.id.client)) => {
+                if let Some(item) = self.store.blocks.get_item(ptr) {
+                    !item.is_deleted()
+                } else {
+                    false
+                }
+            }
+            _ => false,
+        };
+        if trigger {
+            let e = self.changed.entry(parent.ptr.clone()).or_default();
+            e.insert(parent_sub.cloned());
+        }
+    }
 }
 
 impl<'a> Drop for Transaction<'a> {
