@@ -1,35 +1,42 @@
-use crate::block::{BlockPtr, Item, ItemContent, ItemPosition};
-use crate::types::{Inner, TypePtr};
+use crate::block::{ItemContent, ItemPosition, Prelim};
+use crate::types::{Branch, BranchRef, Entries, TypePtr, Value, TYPE_REFS_MAP};
 use crate::*;
 use lib0::any::Any;
-use std::cell::RefCell;
 use std::collections::HashMap;
-use std::rc::Rc;
 
-pub struct Map(Rc<RefCell<Inner>>);
+/// Collection used to store key-value entries in an unordered manner. Keys are always represented
+/// as UTF-8 strings. Values can be any value type supported by Yrs: JSON-like primitives as well as
+/// shared data types.
+///
+/// In terms of conflict resolution, [Map] uses logical last-write-wins principle, meaning the past
+/// updates are automatically overridden and discarded by newer ones, while concurrent updates made
+/// by different peers are resolved into a single value using document id seniority to establish
+/// order.
+#[derive(Debug, Clone, Eq, PartialEq)]
+pub struct Map(BranchRef);
 
 impl Map {
-    pub fn new(inner: Rc<RefCell<Inner>>) -> Self {
-        Map(inner)
-    }
+    /// Converts all entries of a current map into JSON-like object representation.
     pub fn to_json(&self, txn: &Transaction<'_>) -> Any {
-        Self::to_json_inner(&*self.0.borrow(), txn)
-    }
-
-    pub(crate) fn to_json_inner(inner: &Inner, txn: &Transaction<'_>) -> Any {
+        let inner = self.0.as_ref();
         let mut res = HashMap::new();
         for (key, ptr) in inner.map.iter() {
             if let Some(item) = txn.store.blocks.get_item(ptr) {
                 if !item.is_deleted() {
-                    let value = item.content.get_content_last(txn).unwrap_or(Any::Null);
-                    res.insert(key.clone(), value);
+                    let any = if let Some(value) = item.content.get_content_last(txn) {
+                        value.to_json(txn)
+                    } else {
+                        Any::Null
+                    };
+                    res.insert(key.clone(), any);
                 }
             }
         }
         Any::Map(res)
     }
 
-    pub fn len(&self, txn: &Transaction<'_>) -> usize {
+    /// Returns a number of entries stored within current map.
+    pub fn len(&self, txn: &Transaction<'_>) -> u32 {
         let mut len = 0;
         let inner = self.0.borrow();
         for ptr in inner.map.values() {
@@ -43,31 +50,32 @@ impl Map {
         len
     }
 
-    fn blocks<'a, 'b, 'txn>(&'a self, txn: &'b Transaction<'txn>) -> Blocks<'b, 'txn> {
+    fn entries<'a, 'b, 'txn>(&'a self, txn: &'b Transaction<'txn>) -> Entries<'b, 'txn> {
         let ptr = &self.0.borrow().ptr;
-        Blocks::new(ptr, txn)
+        Entries::new(ptr, txn)
     }
 
+    /// Returns an iterator that enables to traverse over all keys of entries stored within
+    /// current map. These keys are not ordered.
     pub fn keys<'a, 'b, 'txn>(&'a self, txn: &'b Transaction<'txn>) -> Keys<'b, 'txn> {
-        Keys(self.blocks(txn))
+        Keys(self.entries(txn))
     }
 
+    /// Returns an iterator that enables to traverse over all values stored within current map.
     pub fn values<'a, 'b, 'txn>(&'a self, txn: &'b Transaction<'txn>) -> Values<'b, 'txn> {
-        Values(self.blocks(txn))
+        Values(self.entries(txn))
     }
 
-    pub fn iter<'a, 'b, 'txn>(&'a self, txn: &'b Transaction<'txn>) -> Iter<'b, 'txn> {
-        Iter(self.blocks(txn))
+    /// Returns an iterator that enables to traverse over all entries - tuple of key-value pairs -
+    /// stored within current map.
+    pub fn iter<'a, 'b, 'txn>(&'a self, txn: &'b Transaction<'txn>) -> MapIter<'b, 'txn> {
+        MapIter(self.entries(txn))
     }
 
-    pub fn insert<V: Into<ItemContent>>(
-        &self,
-        txn: &mut Transaction<'_>,
-        key: String,
-        value: V,
-    ) -> Option<Any> {
+    /// Inserts a new `value` under given `key` into current map. Returns a value stored previously
+    /// under the same key (if any existed).
+    pub fn insert<V: Prelim>(&self, txn: &mut Transaction, key: String, value: V) -> Option<Value> {
         let previous = self.get(txn, &key);
-
         let pos = {
             let inner = self.0.borrow();
             let left = inner.map.get(&key);
@@ -79,37 +87,26 @@ impl Map {
             }
         };
 
-        txn.create_item(&pos, value.into(), Some(key));
+        txn.create_item(&pos, value, Some(key));
         previous
     }
 
-    pub fn remove(&self, txn: &mut Transaction<'_>, key: &String) -> Option<Any> {
+    /// Removes a stored within current map under a given `key`. Returns that value or `None` if
+    /// no entry with a given `key` was present in current map.
+    pub fn remove(&self, txn: &mut Transaction, key: &str) -> Option<Value> {
         let t = self.0.borrow();
-        let ptr = t.map.get(key)?;
-        let prev = {
-            let item = txn.store.blocks.get_item(ptr)?;
-            if item.is_deleted() {
-                None
-            } else {
-                item.content.get_content_last(txn)
-            }
-        };
-        txn.delete(ptr);
-        prev
+        t.remove(txn, key)
     }
 
-    pub fn get(&self, txn: &Transaction<'_>, key: &String) -> Option<Any> {
+    /// Returns a value stored under a given `key` within current map, or `None` if no entry
+    /// with such `key` existed.
+    pub fn get(&self, txn: &Transaction, key: &str) -> Option<Value> {
         let t = self.0.borrow();
-        let ptr = t.map.get(key)?;
-        let item = txn.store.blocks.get_item(ptr)?;
-        if item.is_deleted() {
-            None
-        } else {
-            item.content.get_content_last(txn)
-        }
+        t.get(txn, key)
     }
 
-    pub fn contains(&self, txn: &Transaction<'_>, key: &String) -> bool {
+    /// Checks if an entry with given `key` can be found within current map.
+    pub fn contains(&self, txn: &Transaction, key: &str) -> bool {
         let t = self.0.borrow();
         if let Some(ptr) = t.map.get(key) {
             if let Some(item) = txn.store.blocks.get_item(ptr) {
@@ -119,67 +116,32 @@ impl Map {
         false
     }
 
+    /// Clears the contents of current map, effectively removing all of its entries.
     pub fn clear(&self, txn: &mut Transaction<'_>) {
         let t = self.0.borrow();
         for (_, ptr) in t.map.iter() {
-            if let Some(item) = txn.store.blocks.get_item(ptr) {
-                if !item.is_deleted() {
-                    item.mark_as_deleted();
-                }
-            }
+            txn.delete(ptr);
         }
     }
 }
 
-struct Blocks<'a, 'txn> {
-    txn: &'a Transaction<'txn>,
-    iter: std::collections::hash_map::Iter<'a, String, BlockPtr>,
-}
+pub struct MapIter<'a, 'txn>(Entries<'a, 'txn>);
 
-impl<'a, 'txn> Blocks<'a, 'txn> {
-    fn new<'b>(ptr: &'b TypePtr, txn: &'a Transaction<'txn>) -> Self {
-        let rc = txn.store.get_type(ptr).unwrap();
-        let cell = rc.as_ref();
-        let iter = unsafe { &*cell.as_ptr() as &'a Inner }.map.iter();
-        Blocks { txn, iter }
-    }
-}
-
-impl<'a, 'txn> Iterator for Blocks<'a, 'txn> {
-    type Item = (&'a String, &'a Item);
-
-    fn next(&mut self) -> Option<Self::Item> {
-        let (mut key, ptr) = self.iter.next()?;
-        let mut block = self.txn.store.blocks.get_item(ptr);
-        loop {
-            match block {
-                Some(item) if !item.is_deleted() => {
-                    break;
-                }
-                _ => {
-                    let (k, ptr) = self.iter.next()?;
-                    key = k;
-                    block = self.txn.store.blocks.get_item(ptr);
-                }
-            }
-        }
-        let item = block.unwrap();
-        Some((key, item))
-    }
-}
-
-pub struct Iter<'a, 'txn>(Blocks<'a, 'txn>);
-
-impl<'a, 'txn> Iterator for Iter<'a, 'txn> {
-    type Item = (&'a String, Vec<Any>);
+impl<'a, 'txn> Iterator for MapIter<'a, 'txn> {
+    type Item = (&'a String, Value);
 
     fn next(&mut self) -> Option<Self::Item> {
         let (key, item) = self.0.next()?;
-        Some((key, item.content.get_content(self.0.txn)))
+        if let Some(content) = item.content.get_content_last(self.0.txn) {
+            Some((key, content))
+        } else {
+            self.next()
+        }
     }
 }
 
-pub struct Keys<'a, 'txn>(Blocks<'a, 'txn>);
+/// An unordered iterator over the keys of a [Map].
+pub struct Keys<'a, 'txn>(Entries<'a, 'txn>);
 
 impl<'a, 'txn> Iterator for Keys<'a, 'txn> {
     type Item = &'a String;
@@ -190,10 +152,11 @@ impl<'a, 'txn> Iterator for Keys<'a, 'txn> {
     }
 }
 
-pub struct Values<'a, 'txn>(Blocks<'a, 'txn>);
+/// Iterator over the values of a [Map].
+pub struct Values<'a, 'txn>(Entries<'a, 'txn>);
 
 impl<'a, 'txn> Iterator for Values<'a, 'txn> {
-    type Item = Vec<Any>;
+    type Item = Vec<Value>;
 
     fn next(&mut self) -> Option<Self::Item> {
         let (_, item) = self.0.next()?;
@@ -201,18 +164,45 @@ impl<'a, 'txn> Iterator for Values<'a, 'txn> {
     }
 }
 
-impl Into<ItemContent> for Map {
-    fn into(self) -> ItemContent {
-        ItemContent::Type(self.0.clone())
+impl From<BranchRef> for Map {
+    fn from(inner: BranchRef) -> Self {
+        Map(inner)
+    }
+}
+
+/// A preliminary map. It can be used to early initialize the contents of a [Map], when it's about
+/// to be inserted into another Yrs collection, such as [Array] or another [Map].
+pub struct PrelimMap<T>(HashMap<String, T>);
+
+impl<T> From<HashMap<String, T>> for PrelimMap<T> {
+    fn from(map: HashMap<String, T>) -> Self {
+        PrelimMap(map)
+    }
+}
+
+impl<T: Prelim> Prelim for PrelimMap<T> {
+    fn into_content(self, _txn: &mut Transaction, ptr: TypePtr) -> (ItemContent, Option<Self>) {
+        let inner = BranchRef::new(Branch::new(ptr, TYPE_REFS_MAP, None));
+        (ItemContent::Type(inner), Some(self))
+    }
+
+    fn integrate(self, txn: &mut Transaction, inner_ref: BranchRef) {
+        let map = Map::from(inner_ref);
+        for (key, value) in self.0 {
+            map.insert(txn, key, value);
+        }
     }
 }
 
 #[cfg(test)]
 mod test {
-    use crate::test_utils::exchange_updates;
-    use crate::types::Map;
-    use crate::{Doc, Transaction};
+    use crate::test_utils::{exchange_updates, run_scenario};
+    use crate::types::{Map, Value};
+    use crate::{Doc, PrelimArray, PrelimMap, Transaction};
     use lib0::any::Any;
+    use rand::distributions::Alphanumeric;
+    use rand::prelude::{SliceRandom, StdRng};
+    use rand::Rng;
     use std::collections::HashMap;
 
     #[test]
@@ -238,25 +228,25 @@ mod test {
         m1.insert(&mut t1, "boolean1".to_owned(), true);
         m1.insert(&mut t1, "boolean0".to_owned(), false);
 
-        let m1m = t1.get_map("y-map");
-        let m1a = t1.get_text("y-text");
-        m1a.insert(&mut t1, 0, "a");
-        m1a.insert(&mut t1, 0, "b");
-        m1m.insert(&mut t1, "y-text".to_owned(), m1a);
+        //let m1m = t1.get_map("y-map");
+        //let m1a = t1.get_text("y-text");
+        //m1a.insert(&mut t1, 0, "a");
+        //m1a.insert(&mut t1, 0, "b");
+        //m1m.insert(&mut t1, "y-text".to_owned(), m1a);
 
         //TODO: YArray within YMap
         fn compare_all(t: &Transaction<'_>, m: &Map) {
             assert_eq!(m.len(&t), 5);
-            assert_eq!(m.get(&t, &"number".to_owned()), Some(Any::Number(1f64)));
-            assert_eq!(m.get(&t, &"boolean0".to_owned()), Some(Any::Bool(false)));
-            assert_eq!(m.get(&t, &"boolean1".to_owned()), Some(Any::Bool(true)));
+            assert_eq!(m.get(&t, &"number".to_owned()), Some(Value::from(1f64)));
+            assert_eq!(m.get(&t, &"boolean0".to_owned()), Some(Value::from(false)));
+            assert_eq!(m.get(&t, &"boolean1".to_owned()), Some(Value::from(true)));
             assert_eq!(
                 m.get(&t, &"string".to_owned()),
-                Some(Any::String("hello Y".to_owned()))
+                Some(Value::from("hello Y"))
             );
             assert_eq!(
                 m.get(&t, &"object".to_owned()),
-                Some(Any::Map({
+                Some(Value::from({
                     let mut m = HashMap::new();
                     let mut n = HashMap::new();
                     n.insert("key2".to_owned(), Any::String("value".to_owned()));
@@ -268,8 +258,8 @@ mod test {
 
         compare_all(&t1, &m1);
 
-        let update = d1.encode_state_as_update(&t1);
-        d2.apply_update(&mut t2, update.as_slice());
+        let update = d1.encode_state_as_update_v1(&t1);
+        d2.apply_update_v1(&mut t2, update.as_slice());
 
         compare_all(&t2, &m2);
     }
@@ -283,19 +273,19 @@ mod test {
         m1.insert(&mut t1, "stuff".to_owned(), "stuffy");
         m1.insert(&mut t1, "null".to_owned(), None as Option<String>);
 
-        let update = d1.encode_state_as_update(&t1);
+        let update = d1.encode_state_as_update_v1(&t1);
 
         let d2 = Doc::with_client_id(2);
         let mut t2 = d2.transact();
 
-        d2.apply_update(&mut t2, update.as_slice());
+        d2.apply_update_v1(&mut t2, update.as_slice());
 
         let m2 = t2.get_map("map");
         assert_eq!(
             m2.get(&t2, &"stuff".to_owned()),
-            Some(Any::String("stuffy".to_owned()))
+            Some(Value::from("stuffy"))
         );
-        assert_eq!(m2.get(&t2, &"null".to_owned()), Some(Any::Null));
+        assert_eq!(m2.get(&t2, &"null".to_owned()), Some(Value::Any(Any::Null)));
     }
 
     #[test]
@@ -311,21 +301,14 @@ mod test {
         m1.insert(&mut t1, "stuff".to_owned(), "c0");
         m2.insert(&mut t2, "stuff".to_owned(), "c1");
 
-        let u1 = d1.encode_state_as_update(&t1);
-        let u2 = d2.encode_state_as_update(&t2);
+        let u1 = d1.encode_state_as_update_v1(&t1);
+        let u2 = d2.encode_state_as_update_v1(&t2);
 
-        d1.apply_update(&mut t1, u2.as_slice());
-        d2.apply_update(&mut t2, u1.as_slice());
+        d1.apply_update_v1(&mut t1, u2.as_slice());
+        d2.apply_update_v1(&mut t2, u1.as_slice());
 
-        assert_eq!(
-            m1.get(&t1, &"stuff".to_owned()),
-            Some(Any::String("c1".to_owned()))
-        );
-
-        assert_eq!(
-            m2.get(&t2, &"stuff".to_owned()),
-            Some(Any::String("c1".to_owned()))
-        );
+        assert_eq!(m1.get(&t1, &"stuff".to_owned()), Some(Value::from("c1")));
+        assert_eq!(m2.get(&t2, &"stuff".to_owned()), Some(Value::from("c1")));
     }
 
     #[test]
@@ -342,10 +325,7 @@ mod test {
         assert_eq!(m1.len(&t1), 2);
 
         // remove 'stuff'
-        assert_eq!(
-            m1.remove(&mut t1, &key1),
-            Some(Any::String("c0".to_owned()))
-        );
+        assert_eq!(m1.remove(&mut t1, &key1), Some(Value::from("c0")));
         assert_eq!(m1.len(&t1), 1);
 
         // remove 'stuff' again - nothing should happen
@@ -353,10 +333,7 @@ mod test {
         assert_eq!(m1.len(&t1), 1);
 
         // remove 'other-stuff'
-        assert_eq!(
-            m1.remove(&mut t1, &key2),
-            Some(Any::String("c1".to_owned()))
-        );
+        assert_eq!(m1.remove(&mut t1, &key2), Some(Value::from("c1")));
         assert_eq!(m1.len(&t1), 0);
     }
 
@@ -377,8 +354,8 @@ mod test {
         let d2 = Doc::with_client_id(2);
         let mut t2 = d2.transact();
 
-        let u1 = d1.encode_state_as_update(&t1);
-        d2.apply_update(&mut t2, u1.as_slice());
+        let u1 = d1.encode_state_as_update_v1(&t1);
+        d2.apply_update_v1(&mut t2, u1.as_slice());
 
         let m2 = t2.get_map("map");
         assert_eq!(m2.len(&t2), 0);
@@ -397,12 +374,10 @@ mod test {
             let mut t1 = d1.transact();
             let mut t2 = d2.transact();
             let mut t3 = d3.transact();
-            let mut t4 = d4.transact();
 
             let m1 = t1.get_map("map");
             let m2 = t2.get_map("map");
             let m3 = t3.get_map("map");
-            let m4 = t4.get_map("map");
 
             m1.insert(&mut t1, "key1".to_owned(), "c0");
             m2.insert(&mut t2, "key1".to_owned(), "c1");
@@ -416,12 +391,10 @@ mod test {
             let mut t1 = d1.transact();
             let mut t2 = d2.transact();
             let mut t3 = d3.transact();
-            let mut t4 = d4.transact();
 
             let m1 = t1.get_map("map");
             let m2 = t2.get_map("map");
             let m3 = t3.get_map("map");
-            let m4 = t4.get_map("map");
 
             m1.insert(&mut t1, "key2".to_owned(), "c0");
             m2.insert(&mut t2, "key2".to_owned(), "c1");
@@ -487,7 +460,7 @@ mod test {
 
             assert_eq!(
                 map.get(&txn, &"stuff".to_owned()),
-                Some(Any::String("c3".to_owned())),
+                Some(Value::from("c3")),
                 "peer {} - map entry resolved to unexpected value",
                 doc.client_id
             );
@@ -505,12 +478,10 @@ mod test {
             let mut t1 = d1.transact();
             let mut t2 = d2.transact();
             let mut t3 = d3.transact();
-            let mut t4 = d4.transact();
 
             let m1 = t1.get_map("map");
             let m2 = t2.get_map("map");
             let m3 = t3.get_map("map");
-            let m4 = t4.get_map("map");
 
             m1.insert(&mut t1, "key1".to_owned(), "c0");
             m2.insert(&mut t2, "key1".to_owned(), "c1");
@@ -552,4 +523,62 @@ mod test {
             );
         }
     }
+
+    fn random_string(rng: &mut StdRng) -> String {
+        let len = rng.gen_range(1, 10);
+        rng.sample_iter(&Alphanumeric)
+            .take(len)
+            .map(char::from)
+            .collect()
+    }
+
+    fn map_transactions() -> [Box<dyn Fn(&mut Doc, &mut StdRng)>; 3] {
+        fn set(doc: &mut Doc, rng: &mut StdRng) {
+            let mut txn = doc.transact();
+            let map = txn.get_map("map");
+            let key = ["one", "two"].choose(rng).unwrap();
+            let value: String = random_string(rng);
+            map.insert(&mut txn, key.to_string(), value);
+        }
+
+        fn set_type(doc: &mut Doc, rng: &mut StdRng) {
+            let mut txn = doc.transact();
+            let map = txn.get_map("map");
+            let key = ["one", "two"].choose(rng).unwrap();
+            if rng.gen_bool(0.5) {
+                map.insert(
+                    &mut txn,
+                    key.to_string(),
+                    PrelimArray::from(vec![1, 2, 3, 4]),
+                );
+            } else {
+                map.insert(
+                    &mut txn,
+                    key.to_string(),
+                    PrelimMap::from({
+                        let mut map = HashMap::default();
+                        map.insert("deepkey".to_owned(), "deepvalue");
+                        map
+                    }),
+                );
+            }
+        }
+
+        fn delete(doc: &mut Doc, rng: &mut StdRng) {
+            let mut txn = doc.transact();
+            let map = txn.get_map("map");
+            let key = ["one", "two"].choose(rng).unwrap();
+            map.remove(&mut txn, key);
+        }
+        [Box::new(set), Box::new(set_type), Box::new(delete)]
+    }
+
+    fn fuzzy(iterations: usize) {
+        run_scenario(0, &map_transactions(), 5, iterations)
+    }
+
+    #[test]
+    fn fuzzy_test_6() {
+        fuzzy(6)
+    }
 }
